/*
	Milyn - Copyright (C) 2003

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License (version 2.1) as published by the Free Software 
	Foundation.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
    
	See the GNU Lesser General Public License for more details:    
	http://www.gnu.org/licenses/lgpl.txt
*/

package org.milyn.delivery.serialize;

import java.io.IOException;
import java.io.Writer;

import org.milyn.container.ContainerRequest;
import org.milyn.delivery.ContentDeliveryUnit;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Comment;
import org.w3c.dom.Element;
import org.w3c.dom.EntityReference;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 * W3C Node Serailization interface definition.
 * <p/>
 * Implementations of this interface are applied to the content during the
 * <a href="../doc-files/delivery-phases.html">Serialisation phase</a>.
 * <p/>
 * This interface allows element targeted (by device, profile, DTD info etc)
 * serialization code to be implemented.  It gives control over how an
 * element is "printed" to the target device.
 * <p/>
 * All implementations must contain a public constructor that takes a {@link org.milyn.cdr.CDRDef} instance as 
 * a parameter.  For this reason condsider extending {@link org.milyn.delivery.serialize.AbstractSerializationUnit} for
 * convienience (your IDE should auto-add the constructor).
 * <p/>
 * Serialization Units are defined in the .cdrl file (in .cdrar file(s)) in the very
 * same way as any other content delivery resource (Assembly Units, Transformation Units, 
 * DTDs etc).
 * <p/>
 * Implementations must be stateless in nature.
 * <p/>
 * Only one Serialisation Unit is applied to each element.  If more than one 
 * Serialization Unit is applicable to a given element, the most specific Serialization 
 * Unit is choosen.  See {@link org.milyn.cdr.CDRDefSortComparator}.
 * <p/>
 * See <a href="../package-summary.html">Delivery Overview</a>.
 * @author tfennelly
 */
public interface SerializationUnit extends ContentDeliveryUnit {

	/**
	 * Write the element start portion; the element name and it's
	 * attributes.
	 * <p/>
	 * EG: &lt;a href="http://www.x.com"&gt;
	 * @param element The element start to write.
	 * @param writer The writer to be written to.
	 * @param containerRequest ContainerRequest instance for the delivery context.
	 * @throws IOException Exception writing output.
	 */
	public abstract void writeElementStart(Element element, Writer writer, ContainerRequest containerRequest) throws IOException;

	/**
	 * Write the element end portion; close the element.
	 * <p/>
	 * EG: &lt;/a&gt;
	 * @param element The element end to write.
	 * @param writer The writer to be written to.
	 * @param containerRequest ContainerRequest instance for the delivery context.
	 * @throws IOException Exception writing output.
	 */
	public abstract void writeElementEnd(Element element, Writer writer, ContainerRequest containerRequest) throws IOException;

	/**
	 * Write element text.
	 * @param text The Text object to write.
	 * @param writer The writer to be written to.
	 * @param containerRequest ContainerRequest instance for the delivery context.
	 * @throws IOException Exception writing output.
	 */
	public abstract void writeElementText(Text text, Writer writer, ContainerRequest containerRequest) throws IOException;
	
	/**
	 * Write element comment.
	 * @param comment The comment o write.
	 * @param writer The writer to be written to.
	 * @param containerRequest ContainerRequest instance for the delivery context.
	 * @throws IOException Exception writing output.
	 */
	public abstract void writeElementComment(Comment comment, Writer writer, ContainerRequest containerRequest) throws IOException;

	/**
	 * Write element entity reference object.
	 * @param entityRef The entity reference to write.
	 * @param writer The writer to be written to.
	 * @param containerRequest ContainerRequest instance for the delivery context.
	 * @throws IOException Exception writing output.
	 */
	public abstract void writeElementEntityRef(EntityReference entityRef, Writer writer, ContainerRequest containerRequest) throws IOException;
	
	/**
	 * Write element CDATA section.
	 * @param cdata The CDATA section to write.
	 * @param writer The writer to be written to.
	 * @param containerRequest ContainerRequest instance for the delivery context.
	 * @throws IOException Exception writing output.
	 */
	public abstract void writeElementCDATA(CDATASection cdata, Writer writer, ContainerRequest containerRequest) throws IOException;

	/**
	 * Write element Node object.
	 * <p/>
	 * Called to write DOM types not covered by the other methods on 
	 * this interface.
	 * @param node The node to write.
	 * @param writer The writer to be written to.
	 * @param containerRequest ContainerRequest instance for the delivery context.
	 * @throws IOException Exception writing output.
	 */
	public abstract void writeElementNode(Node node, Writer writer, ContainerRequest containerRequest) throws IOException;
	
	/**
	 * Write the child elements of the element this SerializationUnit is being applied to.
	 * @return True if the child elements are to be writen, otherwise false.
	 */
	public abstract boolean writeChildElements();
}
